"""
    resources.model
    ~~~~~~~~~~~~~~~

    Single model resource object with responders.

    This can be used directly or sub-classed for handling common
    methods against single models.
"""

import falcon
import goldman
import goldman.signals as signals

from ..resources.base import Resource as BaseResource
from goldman.utils.responder_helpers import (
    find,
    from_rest,
    to_rest_model,
)


def on_delete(resc, req, resp, rid):  # pylint: disable=unused-argument
    """ Delete the single item

    Upon a successful deletion an empty bodied 204
    is returned.
    """

    signals.responder_pre_any.send(resc.model)
    signals.responder_pre_delete.send(resc.model)

    model = find(resc.model, rid)
    goldman.sess.store.delete(model)

    resp.status = falcon.HTTP_204

    signals.responder_post_any.send(resc.model)
    signals.responder_post_delete.send(resc.model)


def on_get(resc, req, resp, rid):
    """ Find the model by id & serialize it back """

    signals.responder_pre_any.send(resc.model)
    signals.responder_pre_find.send(resc.model)

    model = find(resc.model, rid)
    props = to_rest_model(model, includes=req.includes)

    resp.last_modified = model.updated
    resp.serialize(props)

    signals.responder_post_any.send(resc.model)
    signals.responder_post_find.send(resc.model)


def on_patch(resc, req, resp, rid):
    """ Deserialize the payload & update the single item """

    signals.responder_pre_any.send(resc.model)
    signals.responder_pre_update.send(resc.model)

    props = req.deserialize()
    model = find(resc.model, rid)

    from_rest(model, props)
    goldman.sess.store.update(model)

    props = to_rest_model(model, includes=req.includes)
    resp.last_modified = model.updated
    resp.serialize(props)

    signals.responder_post_any.send(resc.model)
    signals.responder_post_update.send(resc.model)


class Resource(BaseResource):
    """ Single item resource & responders """

    DESERIALIZERS = [
        goldman.JsonApiDeserializer,
    ]

    SERIALIZERS = [
        goldman.CsvSerializer,
        goldman.JsonApiSerializer,
    ]

    def __init__(self, model, disable=None):

        self.model = model
        self.rondrs = [on_delete, on_get, on_patch]
        self.rtype = model.RTYPE

        super(Resource, self).__init__(disable)



"""
    resources.s3
    ~~~~~~~~~~~~

    AWS S3 object wrangling resource

    Currently, a multipart/form-data file upload is expected
    on POST.
"""

import falcon
import goldman
import goldman.signals as signals
import time

from ..resources.base import Resource as BaseResource
from goldman.exceptions import ServiceUnavailable
from goldman.utils.error_helpers import abort
from goldman.utils.responder_helpers import find
from goldman.utils.s3_helpers import s3_connect, s3_upload


class Resource(BaseResource):
    """ S3 resource & responders """

    MIMETYPES = []

    DESERIALIZERS = [
        goldman.FormDataDeserializer,
    ]

    SERIALIZERS = [
        goldman.JsonSerializer,
    ]

    def __init__(self, model, **kwargs):

        self.model = model
        self.mimetypes = kwargs.get('mimetypes', self.MIMETYPES)

        # s3 properties
        self.acl = kwargs.get('bucket', 'public-read')
        self.bucket = kwargs.get('bucket', goldman.config.S3_BUCKET)
        self.key = kwargs.get('key', goldman.config.S3_KEY)
        self.secret = kwargs.get('secret', goldman.config.S3_SECRET)

        if not self.bucket:
            raise NotImplementedError('an S3 bucket is required')
        super(Resource, self).__init__()

    @property
    def _s3_rtype(self):
        """ Return the chunk after the last / in the URL

        This will be used to generate the path in S3 & should be
        something stable to group the related objects.
        """

        return goldman.sess.req.path.split('/')[-1]

    def _gen_s3_path(self, model, props):
        """ Return the part of the S3 path based on inputs

        The path will be passed to the s3_upload method &
        will ultimately be merged with the standard AWS S3
        URL.

        An example model type of 'users' with a resource ID
        of 99 & an API endpoint ending with 'photos' will have
        a path generated in the following way:

            users/99/photos/<timestamp>.<extension>

        The timestamp is a high precision timestamp & the
        extension is typically 3 characters & derived in
        the form-data deserializer.
        """

        now = '%.5f' % time.time()

        return '%s/%s/%s/%s.%s' % (model.rtype, model.rid_value,
                                   self._s3_rtype, now, props['file-ext'])

    def on_post(self, req, resp, rid):
        """ Deserialize the file upload & save it to S3

        File uploads are associated with a model of some
        kind. Ensure the associating model exists first &
        foremost.
        """

        signals.responder_pre_any.send(self.model)
        signals.responder_pre_upload.send(self.model)

        props = req.deserialize(self.mimetypes)
        model = find(self.model, rid)

        signals.pre_upload.send(self.model, model=model)

        try:
            conn = s3_connect(self.key, self.secret)
            path = self._gen_s3_path(model, props)
            s3_url = s3_upload(self.acl, self.bucket, conn, props['content'],
                               props['content-type'], path)
        except IOError:
            abort(ServiceUnavailable(**{
                'detail': 'The upload attempt failed unexpectedly',
            }))

        signals.post_upload.send(self.model, model=model, url=s3_url)

        resp.location = s3_url
        resp.status = falcon.HTTP_201

        resp.serialize({'data': {'url': s3_url}})

        signals.responder_post_any.send(self.model)
        signals.responder_post_upload.send(self.model)



"""
    signals
    ~~~~~~~

    All of our blinker signals.
"""

import blinker


"""
Signals for our goldman authentication interfaces. See the
interfaces for additional clarity on just when they are called.
"""

# pylint: disable=invalid-name
pre_authenticate = blinker.signal('pre_authenticate')
post_authenticate = blinker.signal('post_authenticate')


"""
Signals for our goldman model based resource responders.
These should be called first thing by the responders.
"""

responder_pre_any = blinker.signal('responder_pre_any')
responder_post_any = blinker.signal('responder_post_any')

responder_pre_create = blinker.signal('responder_pre_create')
responder_post_create = blinker.signal('responder_post_create')

responder_pre_delete = blinker.signal('responder_pre_delete')
responder_post_delete = blinker.signal('responder_post_delete')

responder_pre_find = blinker.signal('responder_pre_find')
responder_post_find = blinker.signal('responder_post_find')

responder_pre_search = blinker.signal('responder_pre_search')
responder_post_search = blinker.signal('responder_post_search')

responder_pre_update = blinker.signal('responder_pre_update')
responder_post_update = blinker.signal('responder_post_update')


"""
Signals for our file upload resources before & after an
upload for a given model.
"""

responder_pre_upload = blinker.signal('responder_pre_upload')
responder_post_upload = blinker.signal('responder_post_upload')

pre_upload = blinker.signal('pre_upload')
post_upload = blinker.signal('post_upload')


"""
Signals for our base goldman models invoked by the store
during there requisite operations
"""

pre_create = blinker.signal('pre_create')
post_create = blinker.signal('post_create')

pre_delete = blinker.signal('pre_delete')
post_delete = blinker.signal('post_delete')

pre_find = blinker.signal('pre_find')
post_find = blinker.signal('post_find')

pre_save = blinker.signal('pre_save')
post_save = blinker.signal('post_save')

pre_search = blinker.signal('pre_search')
post_search = blinker.signal('post_search')

pre_update = blinker.signal('pre_update')
post_update = blinker.signal('post_update')



"""
    utils.s3_helpers
    ~~~~~~~~~~~~~~~~

    Common helpers for interacting with AWS S3

    These interfaces are mostly used by our S3 resource &
    responders but any app could use these as well.
"""

from boto.s3.connection import S3Connection
from boto.s3.key import Key


def gen_url(bucket, path):
    """ Given a path generate the S3 url

    We do this to avoid a round-trip to the S3 API. This
    function may not scale well for the complexities of
    future S3 interactions so be mindful.

    :param key:
        S3 key as generated in our upload() function

    :return:
        A string URL to the object
    """

    return '//s3.amazonaws.com/%s/%s' % (bucket, path)


def s3_connect(key, secret):
    """ Create an S3 connection object using boto """

    return S3Connection(key, secret)


# pylint: disable=too-many-arguments
def s3_upload(acl, bucket, conn, content, content_type, path):
    """ Store an object in our an S3 bucket.

    :param acl:
        S3 ACL for the object
    :param bucket:
        S3 bucket to upload to
    :param content:
        a string representation of the object to upload
    :param content_type:
        a string MIMETYPE of the object that S3 should
        be informed of
    :param path:
        an object specific portion of the S3 key name
        to be passed to gen_url to generate the the location
        in S3 of the new object

    :raise:
        IOError on any failure
    :return:
        S3 generated URL of the uploaded object
    """

    # obj is the object that will be uploaded
    obj = Key(conn.get_bucket(bucket))
    obj.content_type = content_type
    obj.key = path

    obj.set_contents_from_string(content)
    obj.set_acl(acl)

    return gen_url(bucket, path)



"""
    postgres.connect
    ~~~~~~~~~~~~~~~~

    A convenience class to assist in creation of the
    psycopg2 database object.
"""

import goldman
import psycopg2
import psycopg2.extras


class Connect(object):
    """ Custom psycopg2 connection object by the start """

    def __init__(self):

        self._conn = None

    @property
    def config(self):
        """ Return the configuration as needed """

        return goldman.config.PG_URL

    def connect(self):
        """ Construct the psycopg2 connection instance

        :return: psycopg2.connect instance
        """

        if self._conn:
            return self._conn

        self._conn = psycopg2.connect(
            self.config,
            cursor_factory=psycopg2.extras.RealDictCursor,
        )

        self._conn.set_session(autocommit=True)
        psycopg2.extras.register_hstore(self._conn)

        return self._conn



"""
    queryparams.page
    ~~~~~~~~~~~~~~~~

    Paginate resources according to limit offset criteria
    according to.

        jsonapi.org/format/#fetching-pagination

    Only two query params are supported for pagination:

        page[limit]: number of resources to return
        page[offset]: number of resources to skip
"""

import goldman

from goldman.exceptions import InvalidQueryParams


LINK = 'jsonapi.org/format/#fetching-pagination'
PARAM = 'page'


class Paginator(object):
    """ Pagination object

    An instance of this object can be used to express a
    pagination strategy.

    :param limit:
        Integer value of resources to limit a result
        set to
    :param offet:
        Integer value of resources to skip
    """

    def __init__(self, limit, offset):

        self.limit = self._cast_page(limit)
        self.offset = self._cast_page(offset)
        self.total = 0

    def __eq__(self, other):

        try:
            return self.limit == other.limit and \
                self.offset == other.offset
        except AttributeError:
            return False

    def __repr__(self):

        name = self.__class__.__name__

        return '%s(%s, %s)' % (name, self.limit, self.offset)

    def __str__(self):

        return '%s, %s' % (self.limit, self.offset)

    @property
    def current(self):
        """ Generate query parameters for the current page """

        return 'page[offset]=%s&page[limit]=%s' % (self.offset, self.limit)

    @property
    def first(self):
        """ Generate query parameters for the first page """

        if self.total and self.limit < self.total:
            return 'page[offset]=0&page[limit]=%s' % self.limit

    @property
    def last(self):
        """ Generate query parameters for the last page """

        if self.limit > self.total:
            return None
        elif self.offset >= self.total:
            return None
        else:
            offset = (self.total / self.limit) * self.limit
            limit = self.total - offset
            return 'page[offset]=%s&page[limit]=%s' % (offset, limit)

    @property
    def more(self):
        """ Generate query parameters for the next page """

        if self.offset + self.limit + self.limit >= self.total:
            return self.last
        else:
            offset = self.offset + self.limit
            return 'page[offset]=%s&page[limit]=%s' % (offset, self.limit)

    @property
    def prev(self):
        """ Generate query parameters for the prev page """

        if self.total:
            if self.offset - self.limit - self.limit < 0:
                return self.first
            else:
                offset = self.offset - self.limit
                return 'page[offset]=%s&page[limit]=%s' % (offset, self.limit)

    @staticmethod
    def _cast_page(val):
        """ Convert the page limit & offset into int's & type check """

        try:
            val = int(val)
            if val < 0:
                raise ValueError
            return val
        except (TypeError, ValueError):
            raise ValueError


def init(req, model):  # pylint: disable=unused-argument
    """ Determine the pagination preference by query parameter

    Numbers only, >=0, & each query param may only be
    specified once.

    :return: Paginator object
    """

    limit = req.get_param('page[limit]') or goldman.config.PAGE_LIMIT
    offset = req.get_param('page[offset]') or 0

    try:
        return Paginator(limit, offset)
    except ValueError:
        raise InvalidQueryParams(**{
            'detail': 'The page[\'limit\'] & page[\'offset\'] query '
                      'params may only be specified once each & must '
                      'both be an integer >= 0.',
            'links': 'jsonapi.org/format/#fetching-pagination',
            'parameter': 'page',
        })



""" PyPI setup.cfg """

from setuptools import find_packages, setup

setup(
    name='goldman',
    packages=find_packages(),
    version='0.1.7',
    description='An opinionated WSGI web framework',
    author='Sassoo',
    author_email='noreply@devnull.seriously',
    url='https://github.com/sassoo/goldman',
    license='MIT',
    download_url='https://github.com/sassoo/goldman/tarball/0.1.7',
    keywords=['framework', 'http', 'jsonapi', 'rest', 'web', 'wsgi'],
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Topic :: Internet :: WWW/HTTP :: WSGI',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
    ],
    install_requires=[
        'blinker',
        'boto',
        'falcon',
        'phonenumbers',
        'psycopg2',
        'schematics',
        'us',
    ],
)



"""
    resources.s3_model_image
    ~~~~~~~~~~~~~~~~~~~~~~~~

    AWS S3 image upload

    Uses the base s3_model resource but defines the allowed
    mimetypes when uploading.
"""

from ..resources.s3_model import Resource as S3ModelResource


class Resource(S3ModelResource):
    """ S3 model image resource & responders """

    MIMETYPES = ['image/gif', 'image/jpeg', 'image/jpg', 'image/png']



"""
    goldman
    ~~~~~~~

    This is where we initialize the goldman God object.
"""

import threading

sess = threading.local()  # pylint: disable=invalid-name


from goldman.mimetypes import *
from goldman.models import *
from goldman.middleware import *
from goldman.deserializers import *
from goldman.serializers import *
from goldman.resources import *


from goldman.config import Config

config = Config()  # pylint: disable=invalid-name


from goldman.api import *



"""
    types.resource
    ~~~~~~~~~~~~~~

    schematics resource type, type

    This is like a contant type except it does allow being
    written to. Validation then confirms the written value
    is the constant value.
"""

from schematics.exceptions import ValidationError
from schematics.types import StringType


class Type(StringType):
    """ Resource type of the underlying model """

    MESSAGES = {
        'rtype': 'value must be {0}, not {1}'
    }

    def __init__(self, rtype, **kwargs):

        self.rtype = rtype
        super(Type, self).__init__(**kwargs)

    def to_native(self, value, context=None):
        """ Schematics deserializer override """

        if not value:
            return self.rtype
        return value

    def validate_rtype(self, value):
        """ Schematics validator

        The provided rtype must match that of the
        underlying model.

        :return: str
        """

        if not value or value != self.rtype:
            msg = self.messages['rtype'].format(self.rtype, value)
            raise ValidationError(msg)
        return value



"""
    resources.oauth_ropc
    ~~~~~~~~~~~~~~~~~~~~

    OAuth2 Resource Owner Password Credentials Grant resource
    object with responders.

    This resource should be used to accept access_token requests
    according to RFC 6749 section 4.3:

        tools.ietf.org/html/rfc6749#section-4.3

    The resource requires a callable to be passed in as the
    auth_creds property which will be given a username &
    password. The callable should return a tuple of:

        (model, token)

    The model will be assigned to the goldman.sess.login
    propery.

    Returning a string will be interpreted as an error &
    a RFC 6749 compliant error response will be sent with
    the error message as the error_description field in
    the response.

    As an example, if you had a model class named Login with
    a method named auth it could look like:

        class Login(...):
            ...
            def auth(self, username, password):
                login = get_login(username)

                if not login:
                    return 'Invalid username'
                elif login.locked_out:
                    return 'Account has been disabled'
                elif not login.validate_password(password):
                    return 'Invalid password'
                else:
                    return login, token
"""

import falcon
import goldman
import goldman.signals as signals

from goldman.exceptions import AuthRejected
from ..resources.base import Resource as BaseResource


class Resource(BaseResource):
    """ OAuth2 Resource Owner Password Credentials Grant resource """

    DESERIALIZERS = [
        goldman.FormUrlEncodedDeserializer,
    ]

    SERIALIZERS = [
        goldman.JsonSerializer,
    ]

    def __init__(self, auth_creds):

        self.auth_creds = auth_creds
        super(Resource, self).__init__()

    @property
    def _realm(self):
        """ Return a string representation of the authentication realm """

        return 'Bearer realm="%s"' % goldman.config.AUTH_REALM

    def on_post(self, req, resp):
        """ Validate the access token request for spec compliance

        The spec also dictates the JSON based error response
        on failure & is handled in this responder.
        """

        signals.pre_authenticate.send()

        grant_type = req.get_param('grant_type')
        password = req.get_param('password')
        username = req.get_param('username')

        # errors or not, disable client caching along the way
        # per the spec
        resp.disable_caching()

        if not grant_type or not password or not username:
            resp.status = falcon.HTTP_400
            resp.serialize({
                'error': 'invalid_request',
                'error_description': 'A grant_type, username, & password '
                                     'parameters are all required when '
                                     'requesting an OAuth access_token',
                'error_uri': 'tools.ietf.org/html/rfc6749#section-4.3.2',
            })
        elif grant_type != 'password':
            resp.status = falcon.HTTP_400
            resp.serialize({
                'error': 'unsupported_grant_type',
                'error_description': 'The grant_type parameter MUST be set '
                                     'to "password" not "%s"' % grant_type,
                'error_uri': 'tools.ietf.org/html/rfc6749#section-4.3.2',
            })
        else:
            try:
                login, token = self.auth_creds(username, password)
                goldman.sess.login = login
                resp.serialize({
                    'access_token': token,
                    'token_type': 'Bearer',
                })
                signals.post_authenticate.send()
            except AuthRejected as exc:
                resp.status = falcon.HTTP_401
                resp.set_header('WWW-Authenticate', self._realm)
                resp.serialize({
                    'error': 'invalid_client',
                    'error_description': exc.detail,
                })



"""
    stores.base
    ~~~~~~~~~~~

    Base store object.

    This should be sub-classed by other backend specific
    stores. The base store is used as an identity map & caching
    interface.
"""


class Cache(object):
    """ A very simple dictionary cache

    This can be used to store any arbitrary data within the
    context of a single request.
    """

    def __init__(self):

        self._cache = {}

    def get(self, key, bucket):
        """ Get a cached item by key

        If the cached item isn't found the return None.
        """

        try:
            return self._cache[bucket][key]
        except (KeyError, TypeError):
            return None

    def set(self, key, val, bucket):
        """ Set a cached item by key

        WARN: Regardless if the item is already in the cache,
              it will be udpated with the new value.
        """

        if bucket not in self._cache:
            self._cache[bucket] = {}

        self._cache[bucket][key] = val


class Store(object):
    """ Base resource class """

    def __init__(self):

        self.cache = Cache()

    def create(self, model):
        """ Create a new model """

        raise NotImplementedError

    def delete(self, model):
        """ Create an existing model """

        raise NotImplementedError

    def find(self, model, key, val):
        """ Find an existing model """

        raise NotImplementedError

    def query(self, query, param=None):
        """ Perform a store based query """

        raise NotImplementedError

    def search(self, model, **kwargs):
        """ Search for specific kinds of models """

        raise NotImplementedError

    def update(self, model):
        """ Modify an existing model """

        raise NotImplementedError



"""
    validators
    ~~~~~~~~~~

    Adhoc custom schematics type validators to be shared
    across any models or types.
"""

from schematics.exceptions import ValidationError
from uuid import UUID


def validate_int(value):
    """ Integer validator """

    if value and not isinstance(value, int):
        try:
            int(str(value))
        except (TypeError, ValueError):
            raise ValidationError('not a valid number')
    return value


def validate_uuid(value):
    """ UUID 128-bit validator """

    if value and not isinstance(value, UUID):
        try:
            return UUID(str(value), version=4)
        except (AttributeError, ValueError):
            raise ValidationError('not a valid UUID')
    return value



"""
    deserializers
    ~~~~~~~~~~~~~

    All deserializers supported by our API
"""

from ..deserializers.base import Deserializer as BaseDeserializer
from ..deserializers.comma_sep import Deserializer as CsvDeserializer
from ..deserializers.form_data import Deserializer as FormDataDeserializer
from ..deserializers.form_urlencoded import (
    Deserializer as FormUrlEncodedDeserializer
)
from ..deserializers.json_7159 import Deserializer as JsonDeserializer
from ..deserializers.jsonapi import Deserializer as JsonApiDeserializer


DESERIALIZERS = [
    BaseDeserializer,
    CsvDeserializer,
    FormDataDeserializer,
    FormUrlEncodedDeserializer,
    JsonDeserializer,
    JsonApiDeserializer,
]
